from __future__ import annotations

import json
from typing import TYPE_CHECKING, Any

from ..utils import MISSING
from .base_object import ToMessageBase

if TYPE_CHECKING:
    from .results import Result

__all__ = (
    "ErrorResponse",
    "QueryResponse",
    "ExecuteResponse",
)


class BaseResponse(ToMessageBase):
    r"""This represents a response to flow.

    .. WARNING::
        This class is NOT to be used as is. Use one of it's subclasses instead.
    """

    def to_message(self, id: int) -> bytes:
        return (
            json.dumps(
                {
                    "jsonrpc": "2.0",
                    "result": self.to_dict(),
                    "id": id,
                }
            )
            + "\r\n"
        ).encode()


class ErrorResponse(BaseResponse):
    r"""This represents an error sent to or from flow.

    Attributes
    --------
    code: :class:`int`
        The error code for the error
    message: :class:`str`
        The error's message
    data: Optional[Any]
        Any extra data
    """

    __slots__ = "code", "message", "data"

    def __init__(self, code: int, message: str, data: Any | None = None):
        self.code = code
        self.message = message
        self.data = data

    def to_dict(self) -> dict:
        data = self.data
        if isinstance(data, Exception):
            data = f"{data}"
        return {"code": self.code, "message": self.message, "data": data}

    @classmethod
    def from_dict(cls: type[ErrorResponse], data: dict[str, Any]) -> ErrorResponse:
        return cls(code=data["code"], message=data["message"], data=data["data"])

    @classmethod
    def internal_error(cls: type[ErrorResponse], data: Any = None) -> ErrorResponse:
        return cls(code=-32603, message="Internal error", data=data)


class QueryResponse(BaseResponse):
    r"""This response represents the response from search handler's callbacks and context menus. See the :ref:`search handler section <search_handlers>` for more information about using search handlers.

    Attributes
    --------
    results: list[:class:`~flogin.jsonrpc.results.Result`]
        The results to be sent as the result of the query
    settings_changes: dict[:class:`str`, Any]
        Any changes to be made to the plugin's settings.
    debug_message: :class:`str`
        A debug message if you want
    """

    __slots__ = "results", "settings_changes", "debug_message"
    __jsonrpc_option_names__ = {
        "settings_changes": "SettingsChange",
        "debug_message": "debugMessage",
        "results": "result",
    }

    def __init__(
        self,
        results: list[Result],
        settings_changes: dict[str, Any] | None = None,
        debug_message: str = MISSING,
    ):
        self.results = results
        self.settings_changes = settings_changes or {}
        self.debug_message = debug_message or ""


class ExecuteResponse(BaseResponse):
    r"""This response is a generic response for jsonrpc requests, most notably result callbacks.

    Attributes
    --------
    hide: :class:`bool`
        Whether to hide the flow menu after execution or not
    """

    __slots__ = ("hide",)

    def __init__(self, hide: bool = True):
        self.hide = hide



import json
from typing import Any, Self

__all__ = ("Base",)


class Base:
    __slots__ = ()
    __jsonrpc_option_names__: dict[str, str] = {}

    def to_dict(self) -> dict:
        foo = {}
        for name in self.__slots__:
            item = getattr(self, name)
            if isinstance(item, Base):
                item = item.to_dict()
            elif item and isinstance(item, list) and isinstance(item[0], Base):
                item = [item.to_dict() for item in item]
            foo[self.__jsonrpc_option_names__.get(name, name)] = item
        return foo

    @classmethod
    def from_dict(cls: type[Self], data: dict[str, Any]) -> Self:
        raise RuntimeError("This should be overriden")

    def __repr__(self) -> str:
        args = []
        for item in self.__slots__:
            args.append(f"{item}={getattr(self, item)!r}")
        return f"<{self.__class__.__name__} {' '.join(args)}>"


class ToMessageBase(Base):
    def to_message(self, id: int) -> bytes:
        return (json.dumps(self.to_dict()) + "\r\n").encode()



import functools
import logging
import logging.handlers
from functools import _make_key as make_cached_key
from inspect import isasyncgen, iscoroutine
from inspect import signature as _signature
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncIterable,
    Awaitable,
    Callable,
    Coroutine,
    TypeVar,
)

Coro = TypeVar("Coro", bound=Callable[..., Coroutine[Any, Any, Any]])
T = TypeVar("T")

LOG = logging.getLogger(__name__)


class _cached_property:
    def __init__(self, function) -> None:
        self.function = function
        self.__doc__ = getattr(function, "__doc__")

    def __get__(self, instance, owner):
        if instance is None:
            return self

        value = self.function(instance)
        setattr(instance, self.function.__name__, value)

        return value


if TYPE_CHECKING:
    from functools import cached_property as cached_property
else:
    cached_property = _cached_property

__all__ = ("setup_logging", "coro_or_gen", "MISSING", "cached_property", "cached_coro")

def copy_doc(original: Callable[..., Any]) -> Callable[[T], T]:
    def decorator(overridden: T) -> T:
        overridden.__doc__ = original.__doc__
        overridden.__signature__ = _signature(original)  # type: ignore
        return overridden

    return decorator

class _MissingSentinel:
    """A type safe sentinel used in the library to represent something as missing. Used to distinguish from ``None`` values."""

    def __bool__(self) -> bool:
        return False

    def __eq__(self, other: Any) -> bool:
        return False

    def __repr__(self) -> str:
        return "..."


MISSING: Any = _MissingSentinel()


def cached_coro(coro: Coro) -> Coro:
    r"""A decorator to cache a coro's contents based on the passed arguments. This is provided to cache search results.

    .. NOTE::
        The arguments passed to the coro must be hashable.

    Example
    --------
    .. code-block:: python3

        @plugin.search()
        @utils.cached_coro
        async def handler(query):
            ...
    """

    cache = {}

    @functools.wraps(coro)
    async def inner(*args, **kwargs):
        key = make_cached_key(args, kwargs, False)
        try:
            return cache[key]
        except KeyError:
            cache[key] = await coro(*args, **kwargs)
            return cache[key]

    return inner  # type: ignore


def setup_logging(*, formatter: logging.Formatter | None = None) -> None:
    r"""Sets up flogin's default logger.

    Parameters
    ----------
    formatter: Optional[:class:`logging.Formatter`]
        The formatter to use, incase you don't want to use the default file formatter.
    """

    level = logging.DEBUG

    handler = logging.handlers.RotatingFileHandler(
        "flogin.log", maxBytes=1000000, encoding="UTF-8"
    )

    if formatter is None:
        dt_fmt = "%Y-%m-%d %H:%M:%S"
        formatter = logging.Formatter(
            "[{asctime}] [{levelname:<8}] {name}: {message}", dt_fmt, style="{"
        )

    logger = logging.getLogger()
    handler.setFormatter(formatter)
    logger.setLevel(level)
    logger.addHandler(handler)


async def coro_or_gen(coro: Awaitable[T] | AsyncIterable[T]) -> list[T] | T:
    """|coro|

    Executes an AsyncIterable or a Coroutine, and returns the result

    Parameters
    -----------
    coro: :class:`typing.Awaitable` | :class:`typing.AsyncIterable`
        The coroutine or asynciterable to be ran

    Raises
    --------
    TypeError
        Neither a :class:`typing.Coroutine` or an :class:`typing.AsyncIterable` was passed

    Returns
    --------
    Any
        Whatever was given from the :class:`typing.Coroutine` or :class:`typing.AsyncIterable`.
    """

    if iscoroutine(coro):
        return await coro
    elif isasyncgen(coro):
        return [item async for item in coro]
    else:
        raise TypeError(f"Not a coro or gen: {coro!r}")



__title__ = "flogin"
__author__ = "cibere"
__version__ = "0.0.4"


from typing import Literal, NamedTuple

from .conditions import *
from .errors import *
from .jsonrpc import *
from .plugin import *
from .query import *
from .search_handler import *
from .settings import *


class VersionInfo(NamedTuple):
    major: int
    minor: int
    micro: int
    releaselevel: Literal["alpha", "beta", "candidate", "final"]


version_info: VersionInfo = VersionInfo(major=0, minor=0, micro=4, releaselevel="final")

del NamedTuple, Literal, VersionInfo



__all__ = ("JsonRPCException", "JsonRPCVersionMismatch")


class JsonRPCException(Exception):
    r"""This is a base class which represents errors with the jsonrpc client"""


class JsonRPCVersionMismatch(JsonRPCException):
    r"""This is raised when the jsonrpc client receives a jsonrpc version which it isn't looking for

    Attributes
    --------
    expected: :class:`str`
        The Json RPC version it expected
    received: :class:`str`
        The Json RPC version it received
    """

    def __init__(self, expected: str, received: str) -> None:
        super().__init__(
            f"Expected to get version {expected}, but got {received} instead."
        )
        self.expected = expected
        self.received = received



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Any, Awaitable, Literal

from .base import Base, add_prop

if TYPE_CHECKING:
    from .client import FlowLauncherAPI

__all__ = ("PluginMetadata",)


class PluginMetadata(Base):
    r"""This class represents a plugin's metadata

    Attributes
    --------
    id: :class:`str`
        The plugin's ID
    name: :class:`str`
        The plugin's name
    author: :class:`str`
        The name of the plugin's author
    version: :class:`str`
        The current version of the plugin
    language: :class:`str`
        The language that the plugin is written in. Possible values: "csharp", "executable", "fsharp", "python", "javascript", "typescript", "python_v2", "executable_v2", "javascript_v2", "typescript_v2".
    description: :class:`str`
        The plugin's description
    website: :class:`str`
        A link to the plugin's website
    disabled: :class:`bool`
        Whether the plugin is disabled or not
    directory: :class:`str`
        The path to the plugin's directory
    keywords: list[:class:`str`]
        A list of the plugin's keywords
    main_keyword: :class:`str`
        The plugin's main keyword
    """

    def __init__(self, data: dict[str, Any], flow_api: FlowLauncherAPI) -> None:
        super().__init__(data)
        self._flow_api = flow_api

    id: str = add_prop("id")
    name: str = add_prop("name")
    author: str = add_prop("author")
    version: str = add_prop("version")
    language: Literal[
        "csharp",
        "executable",
        "fsharp",
        "python",
        "javascript",
        "typescript",
        "python_v2",
        "executable_v2",
        "javascript_v2",
        "typescript_v2",
    ] = add_prop("language")
    description: str = add_prop("description")
    website: str = add_prop("website")
    disabled: bool = add_prop("disabled")
    directory: str = add_prop("pluginDirectory")
    keywords: list[str] = add_prop("actionKeywords")
    main_keyword: str = add_prop("actionKeyword")

    @property
    def executable(self) -> Path:
        r"""The path to the plugin's executable file"""
        return Path(self._data["executeFilePath"]).absolute()

    @property
    def icon(self) -> Path:
        r"""The path to the plugin's icon file"""
        return Path(self._data["icoPath"]).absolute()

    def add_keyword(self, keyword: str) -> Awaitable[None]:
        r"""|coro|

        Registers a new keyword with flow for the plugin.

        This is a shortcut to :func:`~flogin.flow_api.client.FlowLauncherAPI.add_keyword`

        Parameters
        --------
        keyword: :class:`str`
            The keyword to be added

        Returns
        --------
        None
        """

        return self._flow_api.add_keyword(self.id, keyword)

    def remove_keyword(self, keyword: str) -> Awaitable[None]:
        """|coro|

        Removes a keyword from the plugin.

        This is a shortcut to :func:`~flogin.flow_api.client.FlowLauncherAPI.remove_keyword`

        Parameters
        --------
        keyword: :class:`str`
            The keyword to be removed

        Returns
        --------
        None
        """

        return self._flow_api.remove_keyword(self.id, keyword)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, AsyncIterable, Callable, Coroutine

if TYPE_CHECKING:
    from .query import Query
else:
    Query = Any

SearchHandlerCallbackReturns = Coroutine[Any, Any, Any] | AsyncIterable[Any]
SearchHandlerCallback = Callable[[Query], SearchHandlerCallbackReturns]
SearchHandlerCondition = Callable[[Query], bool]
RawSettings = dict[str, Any]



from inspect import getmembers
from typing import Any


def add_prop(name: str) -> Any:
    return property(lambda self: self._data[name])


class Base:
    __slots__ = ("_data", "__repr_attributes__")

    def __init__(self, data: dict[str, Any]) -> None:
        self._data = data
        self.__repr_attributes__ = [
            entry[0]
            for entry in getmembers(
                self.__class__, lambda other: isinstance(other, property)
            )
        ]

    def __repr__(self) -> str:
        args = []
        for item in self.__repr_attributes__:
            args.append(f"{item}={getattr(self, item)!r}")
        return f"<{self.__class__.__name__} {' '.join(args)}>"



from __future__ import annotations

import asyncio
import json
import logging
from asyncio.streams import StreamReader, StreamWriter
from typing import TYPE_CHECKING, Any

from .errors import JsonRPCException
from .requests import Request
from .responses import BaseResponse, ErrorResponse

LOG = logging.getLogger(__name__)

if TYPE_CHECKING:
    from ..plugin import Plugin

__all__ = ("JsonRPCClient",)


class JsonRPCClient:
    reader: StreamReader
    writer: StreamWriter

    def __init__(self, plugin: Plugin) -> None:
        self.tasks: dict[int, asyncio.Task] = {}
        self.requests: dict[int, asyncio.Future[Any | ErrorResponse]] = {}
        self._current_request_id = 1
        self.plugin = plugin

    @property
    def request_id(self) -> int:
        self._current_request_id += 1
        return self._current_request_id

    @request_id.setter
    def request_id(self, value: int) -> None:
        self._current_request_id = value

    async def request(
        self, method: str, params: list[object] = []
    ) -> Any | ErrorResponse:
        fut: asyncio.Future[Any | ErrorResponse] = asyncio.Future()
        rid = self.request_id
        self.requests[rid] = fut
        msg = Request(method, rid, params).to_message(rid)
        await self.write(msg, drain=False)
        return await fut

    async def __handle_cancellation(self, id: int) -> None:
        if id in self.tasks:
            task = self.tasks.pop(id)
            success = task.cancel()
            if success:
                LOG.info(f"Successfully cancelled task with id {id!r}")
            else:
                LOG.exception(f"Failed to cancel task with id of {id!r}, task={task!r}")
        else:
            LOG.exception(
                f"Failed to cancel task with id of {id!r}, could not find task."
            )

    async def __handle_result(self, result: dict) -> None:
        rid = result["id"]

        LOG.debug(f"Result: {rid}, {result!r}")
        if rid in self.requests:
            try:
                self.requests.pop(rid).set_result(result)
            except asyncio.InvalidStateError:
                pass
        else:
            LOG.exception(
                f"Result from unknown request given. ID: {rid!r}, result={result!r}"
            )

    async def __handle_error(self, id: int, error: ErrorResponse) -> None:
        if id in self.requests:
            self.requests.pop(id).set_exception(Exception(error))
        else:
            LOG.error(f"cancel with no id found: %d", id)

    async def __handle_notification(self, method: str, params: dict[str, Any]) -> None:
        if method == "$/cancelRequest":
            await self.__handle_cancellation(params["id"])
        else:
            LOG.exception(
                f"Unknown notification method received: {method}",
                exc_info=JsonRPCException("Unknown notificaton method received"),
            )

    async def __handle_request(self, request: dict[str, Any]) -> None:
        method: str = request["method"]
        params: list[Any] = request["params"]
        task = None
        error_handler = "on_error"

        self.request_id = request["id"]

        if method.startswith("flogin.action"):
            slug = method.removeprefix("flogin.action.")
            result = self.plugin._results.get(slug)
            if result:
                callback = result.callback
                error_handler = result.on_error
                result.plugin = self.plugin
                task = self.plugin._schedule_event(
                    callback, method, error_handler=error_handler
                )

        if task is None:
            task = self.plugin.dispatch(method, *params)
            if not task:
                return

        self.tasks[request["id"]] = task
        result = await task

        if isinstance(result, BaseResponse):
            return await self.write(result.to_message(id=request["id"]))
        else:
            return await self.write(
                ErrorResponse.internal_error().to_message(id=request["id"])
            )

    async def start_listening(self, reader: StreamReader, writer: StreamWriter):
        self.reader = reader
        self.writer = writer

        while 1:
            bytes = await reader.readline()
            line = bytes.decode("utf-8")
            if line == "":
                continue

            LOG.info(f"Received line: {line}")
            message = json.loads(line)

            if "id" not in message:
                LOG.debug(f"Received notification: {message!r}")
                asyncio.create_task(
                    self.__handle_notification(message["method"], message["params"])
                )
            elif "method" in message:
                LOG.debug(f"Received request: {message!r}")
                asyncio.create_task(self.__handle_request(message))
            elif "result" in message:
                LOG.debug(f"Received result: {message!r}")
                asyncio.create_task(self.__handle_result(message))
            elif "error" in message:
                LOG.exception(f"Received error: {message!r}")
                asyncio.create_task(
                    self.__handle_error(
                        message["id"], ErrorResponse.from_dict(message["error"])
                    )
                )
            else:
                LOG.exception(
                    f"Unknown message type received",
                    exc_info=JsonRPCException("Unknown message type received"),
                )

    async def write(self, msg: bytes, drain: bool = True) -> None:
        LOG.debug(f"Sending: {msg!r}")
        self.writer.write(msg)
        if drain:
            await self.writer.drain()



from __future__ import annotations

from typing import Any, Generic, TypeVar

T = TypeVar("T")

__all__ = ("Query",)


class Query(Generic[T]):
    r"""This class represents the query data sent from flow launcher

    .. container:: operations

        .. describe:: x == y

            Compare the keywords, text, and is_query values of two query objects.

        .. describe:: hash(x)

            Gets the hash of the query's raw text
    """

    def __init__(self, data: dict[str, Any]) -> None:
        self.__data = data
        self.__search_condition_data: T | None = None

    @property
    def raw_text(self) -> str:
        """:class:`str`: The raw and complete query, which includes the keyword"""
        return self.__data["rawQuery"]

    @property
    def is_requery(self) -> bool:
        """:class:`bool`: Whether the query is a requery or not"""
        return self.__data["isReQuery"]

    @property
    def text(self) -> str:
        """:class:`str`: The actual query, excluding any keywords"""
        return self.__data["search"]

    @property
    def keyword(self) -> str:
        """:class:`str`: The keyword used to initiate the query"""
        return self.__data["actionKeyword"]

    @property
    def condition_data(self) -> T | None:
        """Any | None: If used in a :class:`~flogin.search_handler.SearchHandler`, this attribute will return any extra data that the condition gave."""
        return self.__search_condition_data

    @condition_data.setter
    def condition_data(self, value: T) -> None:
        self.__search_condition_data = value

    def __eq__(self, other: Any) -> bool:
        return (
            isinstance(other, Query)
            and other.raw_text == self.raw_text
            and other.is_requery == self.is_requery
        )

    def __hash__(self) -> int:
        return hash(self.raw_text)



from __future__ import annotations

import logging
import random
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Coroutine,
    Iterable,
    NotRequired,
    Self,
    TypedDict,
    TypeVarTuple,
    Unpack,
)

from ..utils import cached_property, copy_doc
from .base_object import Base
from .responses import ErrorResponse

if TYPE_CHECKING:
    from .._types import SearchHandlerCallbackReturns
    from ..plugin import Plugin
    from .responses import ExecuteResponse

TS = TypeVarTuple("TS")
LOG = logging.getLogger(__name__)

__all__ = ("Result",)


class ResultConstructorArgs(TypedDict):
    title: str
    sub: NotRequired[str | None]
    icon: NotRequired[str | None]
    title_highlight_data: NotRequired[Iterable[int] | None]
    title_tooltip: NotRequired[str | None]
    sub_tooltip: NotRequired[str | None]
    copy_text: NotRequired[str | None]
    score: NotRequired[int | None]


class Result(Base):
    r"""This represents a result that would be returned as a result for a query or context menu.

    For simple useage: create instances of this class as-is.

    For advanced useage (handling clicks and custom context menus), it is recommended to subclass the result object to create your own subclass.

    Subclassing
    ------------
    Subclassing lets you override the following methods: :func:`~flogin.jsonrpc.results.Result.callback` and :func:`~flogin.jsonrpc.results.Result.context_menu`. It also lets you create "universal" result properties (eg: same icon). Example:

    .. code-block:: python3

        class MyResult(Result):
            def __init__(self, title: str) -> None:
                super().__init__(self, title, icon="Images/app.png")

            async def callback(self):
                # handle what happens when the result gets clicked

            async def context_menu(self):
                # add context menu options to this result's context menu

    Attributes
    ----------
    title: :class:`str`
        The title/content of the result
    sub: Optional[:class:`str`]
        The subtitle to be shown.
    icon: Optional[:class:`str`]
        A path to the icon to be shown with the result.
    title_highlight_data: Optional[Iterable[:class:`int`]]
        The highlight data for the title. See the :ref:`FAQ section on highlights <highlights>` for more info.
    title_tooltip: Optional[:class:`str`]
        The text to be displayed when the user hovers over the result's title
    sub_tooltip: Optional[:class:`str`]
        The text to be displayed when the user hovers over the result's subtitle
    copy_text: Optional[:class:`str`]
        This is the text that will be copied when the user does ``CTRL+C`` on the result.
    plugin: :class:`~flogin.plugin.Plugin` | None
        Your plugin instance. This is filled before :func:`~flogin.jsonrpc.results.Result.callback` or :func:`~flogin.jsonrpc.results.Result.context_menu` are triggered.
    """

    def __init__(
        self,
        title: str,
        sub: str | None = None,
        icon: str | None = None,
        title_highlight_data: Iterable[int] | None = None,
        title_tooltip: str | None = None,
        sub_tooltip: str | None = None,
        copy_text: str | None = None,
        score: int | None = None,
    ) -> None:
        self.title = title
        self.sub = sub
        self.icon = icon
        self.title_highlight_data = title_highlight_data
        self.title_tooltip = title_tooltip
        self.sub_tooltip = sub_tooltip
        self.copy_text = copy_text
        self.score = score
        self.plugin: Plugin | None = None

    async def on_error(self, error: Exception) -> ErrorResponse | ExecuteResponse:
        r"""|coro|

        Override this function to add an error response behavior to this result's callback.

        Parameters
        ----------
        error: :class:`Exception`
            The error that occured
        """
        LOG.exception(
            f"Ignoring exception in reuslt callback ({self!r})", exc_info=error
        )
        return ErrorResponse.internal_error(error)

    async def callback(self) -> ExecuteResponse:
        r"""|coro|

        Override this function to add a callback behavior to your result. This method will run when the user clicks on your result.

        Returns
        -------
        :class:`~flogin.jsonrpc.responses.ExecuteResponse`
            A response to flow determining whether or not to hide flow's menu
        """

        return ExecuteResponse(False)

    def context_menu(self) -> SearchHandlerCallbackReturns:
        r"""|coro|

        Override this function to add a context menu behavior to your result. This method will run when the user gets the context menu to your result.

        This method can return/yield almost anything, and flogin will convert it into a list of :class:`~flogin.jsonrpc.results.Result` objects before sending it to flow.

        Returns
        -------
        list[:class:`~flogin.jsonrpc.results.Result`] | :class:`~flogin.jsonrpc.results.Result` | str | Any
            A list of results, an results, or something that can be converted into a list of results.

        Yields
        ------
        :class:`~flogin.jsonrpc.results.Result` | str | Any
            A result object or something that can be converted into a result object.
        """
        ...

    def on_context_menu_error(
        self, error: Exception
    ) -> SearchHandlerCallbackReturns:
        r"""|coro|

        Override this function to add an error response behavior to this result's context menu callback.

        If the error was handled:
            You can return/yield almost anything, and flogin will convert it into a list of :class:`~flogin.jsonrpc.results.Result` objects before sending it to flow.

        If the error was not handled:
            Return a :class:`~flogin.jsonrpc.responses.ErrorResponse` object

        Parameters
        ----------
        error: :class:`Exception`
            The error that occured

        Returns
        -------
        :class:`~flogin.jsonrpc.responses.ErrorResponse` | list[:class:`~flogin.jsonrpc.results.Result`] | :class:`~flogin.jsonrpc.results.Result` | str | Any
            A list of results, an results, or something that can be converted into a list of results.

        Yields
        ------
        :class:`~flogin.jsonrpc.results.Result` | str | Any
            A result object or something that can be converted into a result object.
        """
        ...

    if not TYPE_CHECKING:
        
        @copy_doc(context_menu)
        async def context_menu(self):
            return []

        @copy_doc(on_context_menu_error)
        async def on_context_menu_error(self, error: Exception):
            LOG.exception(
                f"Ignoring exception in result's context menu callback ({self!r})",
                exc_info=error,
            )
            return ErrorResponse.internal_error(error)

    def to_dict(self) -> dict[str, Any]:
        r"""This converts the result into a json serializable dictionary

        Returns
        -------
        dict[:class:`str`, Any]
        """

        x: dict[str, Any] = {
            "title": self.title,
        }
        if self.sub is not None:
            x["subTitle"] = self.sub
        if self.icon is not None:
            x["icoPath"] = self.icon
        if self.title_highlight_data is not None:
            x["titleHighlightData"] = self.title_highlight_data
        if self.title_tooltip is not None:
            x["titleTooltip"] = self.title_tooltip
        if self.sub_tooltip is not None:
            x["subtitleTooltip"] = self.sub_tooltip
        if self.copy_text is not None:
            x["copyText"] = self.copy_text
        if self.callback is not None:
            x["jsonRPCAction"] = {"method": f"flogin.action.{self.slug}"}
        if self.context_menu is not None:
            x["ContextData"] = [self.slug]
        if self.score is not None:
            x["score"] = self.score
        return x

    @classmethod
    def from_dict(cls: type[Self], data: dict[str, Any]) -> Self:
        r"""Creates a Result from a dictionary

        .. NOTE::
            This method does NOT fill the :func:`~flogin.jsonrpc.results.Result.callback` or :func:`~flogin.jsonrpc.results.Result.context_menu` attributes.

        Parameters
        ----------
        data: dict[:class:`str`, Any]
            The valid dictionary that includes the result data

        Raises
        ------
        :class:`KeyError`
            The dictionary did not include the only required field, ``title``.

        Returns
        --------
        :class:`Result`
        """

        return cls(
            title=data["title"],
            sub=data.get("subTitle"),
            icon=data.get("icoPath"),
            title_highlight_data=data.get("titleHighlightData"),
            title_tooltip=data.get("titleTooltip"),
            sub_tooltip=data.get("subtitleTooltip"),
            copy_text=data.get("copyText"),
        )

    @classmethod
    def from_anything(cls: type[Result], item: Any) -> Result:
        if isinstance(item, dict):
            return cls.from_dict(item)
        elif isinstance(item, Result):
            return item
        else:
            return cls(str(item))

    @classmethod
    def create_with_partial(
        cls: type[Result],
        partial_callback: Callable[[], Coroutine[Any, Any, Any]],
        **kwargs: Unpack[ResultConstructorArgs],
    ) -> Result:
        r"""A quick and easy way to create a result with a callback without subclassing.

        .. NOTE::
            This is meant to be used with :class:`~flogin.flow_api.client.FlowLauncherAPI` methods

        Example
        --------
        .. code-block:: python3

            result = Result.create_with_partial(
                functools.partial(
                    plugin.api.show_notification,
                    "notification title",
                    "notification content"
                ),
                title="Result title",
                sub="Result subtitle"
            )

        Parameters
        ----------
        partial_callback: partial :ref:`coroutine <coroutine>`
            The callback wrapped in :obj:`functools.partial`
        kwargs: See allowed kwargs here: :class:`~flogin.jsonrpc.results.Result`
            The args that will be passed to the :class:`~flogin.jsonrpc.results.Result` constructor
        """

        self = cls(**kwargs)
        self.callback = partial_callback
        return self

    @cached_property
    def slug(self) -> str:
        return "".join(
            random.choices(
                "QWERTYUIOPASDFGHJKLZXCVBNMqwertyuiopasdfghjklzxcvbnm1234567890", k=15
            )
        )



from __future__ import annotations

import re
from typing import TYPE_CHECKING, Iterable

if TYPE_CHECKING:
    from ._types import SearchHandlerCondition
    from .query import Query

__all__ = "PlainTextCondition", "RegexCondition", "KeywordCondition", "MultiCondition"


class PlainTextCondition:
    r"""A builtin search condition to check plain text.

    This condition will only run if the query's text is the same as the text given to this condition.
    See the :ref:`search handler section <search_handlers>` for more information about using search handlers and conditions.

    Attributes
    ----------
    text: :class:`str`
        The text to compare the query to
    """

    __slots__ = ("text",)

    def __init__(self, text: str):
        self.text = text

    def __call__(self, query: Query) -> bool:
        return query.text == self.text


class RegexCondition:
    r"""A builtin search condition to check a regex pattern.

    This condition will only run if the query's text is a match to the regex pattern given to this condition.
    See the :ref:`search handler section <search_handlers>` for more information about using search handlers and conditions.

    .. NOTE::
        This condition will set the query's :attr:`~flogin.query.Query.condition_data` attribute to the :class:`re.Match` object.

    Attributes
    ----------
    pattern: :class:`re.Pattern`
        The pattern to check the queries against.
    """

    __slots__ = ("pattern",)

    def __init__(self, pattern: re.Pattern):
        self.pattern = pattern

    def __call__(self, query: Query[re.Match]) -> bool:
        match = self.pattern.match(query.text)
        if match:
            query.condition_data = match
            return True
        return False


class MultiCondition:
    r"""A builtin search condition to check for multiple conditions.

    This condition will only run if all given conditions return ``True``.
    See the :ref:`search handler section <search_handlers>` for more information about using search handlers and conditions.

    .. NOTE::
        This condition will set the query's :attr:`~flogin.query.Query.condition_data` attribute to a dictionary object where the key is the condition and the value is the extra data it provided.

    """

    __slots__ = ("conditions",)

    def __init__(self, conditions: Iterable[SearchHandlerCondition]) -> None:
        self.conditions = conditions

    def __call__(self, query: Query) -> bool:
        condition_data = {}
        for condition in self.conditions:
            if condition(query) is False:
                return False
            condition_data[condition.__name__] = query.condition_data
            query.condition_data = None

        query.condition_data = condition_data
        return True


class KeywordCondition:
    r"""A builtin search condition to check what keyword was used with the query.

    If the :attr:`~flogin.conditions.KeywordCondition.allowed_keywords` attribute is given, the handler will only run if the query's keyword is in the list of allowed keywords.
    If the :attr:`~flogin.conditions.KeywordCondition.disallowed_keywords` attribute is given, the handler will only run if the query's keyword is not in the list of allowed keywords.

    See the :ref:`search handler section <search_handlers>` for more information about using search handlers and conditions.

    Attributes
    ----------
    allowed_keywords: Optional[Iterable[:class:`str`]]
        The allowed keywords
    disallowed_keywords: Optional[Iterable[:class:`str`]]
        The disallowed keywords
    """

    __slots__ = "allowed_keywords", "disallowed_keywords"

    def __init__(
        self,
        *,
        allowed_keywords: Iterable[str] | None = None,
        disallowed_keywords: Iterable[str] | None = None,
    ) -> None:
        if allowed_keywords is None and disallowed_keywords is None:
            raise TypeError(
                "Either the 'allowed_keywords' arg or the 'disallowed_keywords' arg must be given"
            )
        elif allowed_keywords is not None and disallowed_keywords is not None:
            raise TypeError(
                "'allowed_keywords' and 'disallowed_keywords' can not be passed together. Use `MultiCondition` if you would like to achieve it."
            )

        self.allowed_keywords: Iterable[str] | None = allowed_keywords or None
        self.disallowed_keywords: Iterable[str] | None = disallowed_keywords or None

    def __call__(self, query: Query) -> bool:
        if self.allowed_keywords is None and self.disallowed_keywords is not None:
            return query.keyword not in self.disallowed_keywords
        elif self.allowed_keywords is not None and self.disallowed_keywords is None:
            return query.keyword in self.allowed_keywords
        elif self.allowed_keywords is None and self.disallowed_keywords is None:
            raise RuntimeError(
                "'allowed_keywords' and 'disallowed_keywords' are both set to None. How did this happen?"
            )
        elif self.allowed_keywords is not None and self.disallowed_keywords is not None:
            raise RuntimeError(
                "'allowed_keywords' and 'disallowed_keywords' are provided. How did this happen?"
            )
        else:
            raise RuntimeError("How did we get here")



import os

import click, json, uuid


@click.group
def cli(): ...

@cli.group(name="create", help="Create specific files")
def create_group(): ...

@create_group.command(name="plugin.json", help="Creates a new plugin.json file")
def create_file() -> None:
    name = input("Plugin Name\n> ")
    desc = input("Plugin Description\n> ")
    author = input("Author Name\n> ")
    plugin_website = input("Plugin Website\n> ")
    icon_path = input("Icon Path (leave blank for `Images/app.png`)\n> ")
    main_file = input("What file should flow execute to start the plugin? Leave blank for `main.py`\n> ")
    data = {
        "ID": uuid.uuid4(),
        "ActionKeyword": "test",
        "Name": name,
        "Description": desc,
        "Author": author,
        "Version": "0.0.1",
        "Language": "python_v2",
        "Website": plugin_website,
        "IcoPath": icon_path,
        "ExecuteFileName": main_file
    }
    with open("plugin.json", "w") as f:
        json.dump(data, f)

@create_group.group("gh", help="Create files in the .github directory")
def create_gh_file_group(): ...

@create_gh_file_group.group("issue_template", help="Github issue templates")
def create_issue_template_group(): ...

@create_issue_template_group.command(name="bug_report", help="Create a detailed bug report template for github issues")
def create_bug_report_issue_template():
    content = f"""
name: Bug Report
description: Report broken or incorrect behaviour
labels: unconfirmed bug
body:
    - type: markdown
        attributes:
        value: >
            Thanks for taking the time to fill out a bug.

            Please note that this form is for bugs only!
    - type: input
        attributes:
            label: Summary
            description: A simple summary of your bug report
        validations:
            required: true
    - type: textarea
        attributes:
            label: Reproduction Steps
            description: >
                What you did to make it happen.
        validations:
            required: true
    - type: textarea
        attributes:
            label: Minimal Reproducible Code
            description: >
                A short snippet of code that showcases the bug.
        render: python
    - type: textarea
        attributes:
            label: Expected Results
            description: >
                What did you expect to happen?
        validations:
            required: true
    - type: textarea
        attributes:
            label: Actual Results
            description: >
                What actually happened?
        validations:
            required: true
    - type: textarea
        attributes:
            label: Flow Launcher Version
            description: Go into your flow launcher settings, go into the about section, and the version should be at the top.
        validations:
            required: true
    - type: textarea
        attributes:
            label: Python Version/Path
            description: Go into your flow launcher settings, go to the general section, and scroll down until you find the `Python Path` field. Copy and paste the value here.
        validations:
            required: true
    - type: textarea
        attributes:
            label: If applicable, Flow Launcher Log File
            description: Use the `Open Log Location` command with the `System Commands` plugin to open the log file folder, and upload the newest file here.
    - type: textarea
        attributes:
            label: Flogin Log File
            description: Use the `Flow Launcher UserData Folder` command with the `System Commands` plugin to open your userdata folder, go into the `Plugins` folder, then find the plugin and go into it. If the `flogin.log` file exists, upload it here. Otherwise please state that it was not there.
    - type: checkboxes
        attributes:
            label: Checklist
            description: >
                Let's make sure you've properly done due diligence when reporting this issue!
            options:
                - label: I have searched the open issues for duplicates.
                required: true
                - label: I have shown the entire traceback, if possible.
                required: true
                - label: I have removed my token from display, if visible.
                required: true
    - type: textarea
        attributes:
            label: Additional Context
            description: If there is anything else to say, please do so here.
    """.strip()
    if not os.path.isdir(".github"):
        os.mkdir(".github")
    if not os.path.isdir(".github/ISSUE_TEMPLATE"):
        os.mkdir(".github/ISSUE_TEMPLATE")
    
    with open(".github/ISSUE_TEMPLATE/bug_report.yml", "w") as f:
        f.write(content)

@create_gh_file_group.command(name="pr_template", help="Create a basic PR template")
def pr_template():
    content = """
## Summary

<!-- What is this pull request for? Does it fix any issues? -->

## Checklist

<!-- Put an x inside [ ] to check it, like so: [x] -->

- [ ] If code changes were made then they have been tested.
    - [ ] I have updated the documentation to reflect the changes.
- [ ] This PR fixes an issue.
- [ ] This PR adds something new (e.g. new method or parameters).
- [ ] This PR is a breaking change (e.g. methods or parameters removed/renamed)
- [ ] This PR is **not** a code change (e.g. documentation, README, ...)
    """.strip()
    if not os.path.isdir(".github"):
        os.mkdir(".github")
    
    with open(".github/PULL_REQUEST_TEMPLATE.md", "w") as f:
        f.write(content)

@create_gh_file_group.group("workflows", help="Create github workflows")
def create_gh_workflows_group(): ...

@create_gh_workflows_group.command(name="publish_release", help="A standard workflow to publish and release a new version of your plugin")
@click.option(
    "--changelog",
    is_flag=True,
    help="If passed, a `CHANGLOG.txt` file will be created in the root directory. When the workflow gets run, the contents of that file will be used as the release's changelog/description."
)
def create_publish_and_release_workflow(changelog: bool = False):
    content = """
name: Publish and Release

on:
  workflow_dispatch:

jobs:
  publish:
    runs-on: ubuntu-latest
    env:
      python_ver: 3.11
    permissions:
      contents: write

    steps:
      - uses: actions/checkout@v2

      - name: Set up Python ${{ matrix.python-version }}
        uses: actions/setup-python@v2
        with:
          python-version: ${{ matrix.python-version }}

      - name: get version
        id: version
        uses: notiz-dev/github-action-json-property@release
        with: 
          path: 'plugin.json'
          prop_path: 'Version'

      - run: echo ${{steps.version.outputs.prop}} 

      - name: Install dependencies
        run: |
          python -m pip install --upgrade pip
          pip install -r ./requirements.txt -t ./lib
          zip -r ${{ github.event.repository.name }} . -x '*.git*'

      - name: Publish
        if: success()
        uses: softprops/action-gh-release@v2
        with:
          files: '${{ github.event.repository.name }}.zip'
          tag_name: "v${{steps.version.outputs.prop}}"
    """.strip()
    if changelog:
        content += "\n          body_path: 'CHANGELOG.txt'"
        with open("CHANGELOG.txt", "w") as f:
            f.write(f"# v0.0.1")

    if not os.path.isdir(".github"):
        os.mkdir(".github")
    if not os.path.isdir(".github/workflows"):
        os.mkdir(".github/workflows")
    
    with open(".github/workflows/publish_release.yml", "w") as f:
        f.write(content)

if __name__ == "__main__":
    cli()



from __future__ import annotations

__all__ = (
    "PluginException",
    "SettingNotFound",
    "PluginNotInitialized",
)


class PluginException(Exception):
    r"""A class that represents exceptions with your plugin"""


class SettingNotFound(PluginException):
    r"""This gets raised when you try to access a setting key that doesn't exist

    Attributes
    -----------
    name: :class:`str`
        The name of the setting key
    """

    def __init__(self, name: str) -> None:
        super().__init__(f"Setting {name!r} was not found.")
        self.name = name


class PluginNotInitialized(PluginException):
    r"""This is raised when you try to access something that needs data from the initialize method, and it hasn't been called yet."""

    def __init__(self):
        return super().__init__("The plugin hasn't been initialized yet")


class ContextMenuHandlerException(PluginException):
    r"""This is a base class for errors related to context menu handlers."""

    ...


class InvalidContextDataReceived(ContextMenuHandlerException):
    r"""Invalid context menu data was provided"""

    def __init__(self):
        return super().__init__(f"Invalid context menu data received")



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from .errors import SettingNotFound

if TYPE_CHECKING:
    from ._types import RawSettings

__all__ = ("Settings",)


class Settings:
    r"""This class represents the settings that you user has chosen

    .. container:: operations

        .. describe:: x['setting name']

            Get a setting by key similiar to a dictionary

        .. describe:: x['setting name'] = "new value"

            Change a settings value like a dictionary

        .. describe:: x.setting_name

            Get a setting by name like an attribute

        .. describe:: x.setting_name = "new value"

            Change a settings value like an attribute


    Raises
    --------
    :class:`SettingNotFound`
        A setting was not found
    """

    _data: RawSettings
    _changes: RawSettings

    def __init__(self, data: RawSettings) -> None:
        self._data = data
        self._changes = {}

    def __getitem__(self, key: str) -> Any:
        try:
            return self._data[key]
        except KeyError:
            raise SettingNotFound(key) from None

    def __setitem__(self, key: str, value: Any) -> None:
        self._data[key] = value
        self._changes[key] = value

    def __getattribute__(self, name: str) -> Any:
        if name.startswith("_"):
            try:
                return super().__getattribute__(name)
            except AttributeError as e:
                raise AttributeError(
                    f"{e}. Settings that start with an underscore (_) can only be accessed by the __getitem__ method. Ex: settings['_key']"
                ) from None
        return self.__getitem__(name)

    def __setattr__(self, name: str, value: Any) -> None:
        if name.startswith("_"):
            return super().__setattr__(name, value)
        self.__setitem__(name, value)

    def _update(self, data: RawSettings) -> None:
        self._data = data
        self._changes = {}

    def __repr__(self) -> str:
        return f"<Settings current={self._data!r}, pending_changes={self._changes}>"



from __future__ import annotations

from typing import Any

from .base_object import ToMessageBase

__all__ = ("Request",)


class Request(ToMessageBase):
    __slots__ = "method", "id", "params"

    def __init__(self, method: str, id: int, params: list[Any] | None = None):
        self.method = method
        self.id = id
        self.params = params

    def to_dict(self) -> dict:
        x = super().to_dict()
        x["jsonrpc"] = "2.0"
        return x



from .base import Base, add_prop

__all__ = ("FuzzySearchResult",)


class FuzzySearchResult(Base):
    r"""A class which represents the result given from flow launcher to a fuzzy search request

    .. NOTE::
        This is not intended to be a class that you create yourself, use :func:`~flogin.flow_api.client.FlowLauncherAPI.fuzzy_search` instead.

    Attributes
    --------
    score: :class:`int`
        The score of the result
    highlight_data: list[:class:`int`]
        The highlight data included with the result
    search_percision: :class:`int`
        The perision of the result
    """

    score: int = add_prop("score")
    highlight_data: list[int] = add_prop("matchData")
    search_precision: int = add_prop("searchPrecision")



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, Callable, TypeVar

from .jsonrpc import ErrorResponse
from .utils import copy_doc

if TYPE_CHECKING:
    from ._types import SearchHandlerCallbackReturns, SearchHandlerCondition
    from .plugin import Plugin
    from .query import Query

    ErrorHandlerT = TypeVar(
        "ErrorHandlerT",
        bound=Callable[[Query, Exception], SearchHandlerCallbackReturns],
    )

LOG = logging.getLogger(__name__)

__all__ = ("SearchHandler",)


def _default_condition(q: Query) -> bool:
    return True


class SearchHandler:
    r"""This represents a search handler.

    When creating this on your own, the :func:`~flogin.plugin.Plugin.register_search_handler` method can be used to register it.

    See the :ref:`search handler section <search_handlers>` for more information about using search handlers.

    There is a provided decorator to easily create search handlers: :func:`~flogin.plugin.Plugin.search`

    Attributes
    ------------
    condition: :ref:`condition <condition_example>`
        A function which is used to determine if this search handler should be used to handle a given query or not
    plugin: :class:`~flogin.plugin.Plugin` | None
        Your plugin instance. This is filled before :func:`~flogin.search_handler.SearchHandler.callback` is triggered.
    """

    def __init__(
        self,
        condition: SearchHandlerCondition | None = None,
    ) -> None:
        if condition is None:
            condition = _default_condition

        self.condition = condition
        self.plugin: Plugin | None = None

    def callback(self, query: Query) -> SearchHandlerCallbackReturns:
        r"""|coro|

        Override this function to add the search handler behavior you want for the set condition.

        This method can return/yield almost anything, and flogin will convert it into a list of :class:`~flogin.jsonrpc.results.Result` objects before sending it to flow.

        Returns
        -------
        list[:class:`~flogin.jsonrpc.results.Result`] | :class:`~flogin.jsonrpc.results.Result` | str | Any
            A list of results, an results, or something that can be converted into a list of results.

        Yields
        ------
        :class:`~flogin.jsonrpc.results.Result` | str | Any
            A result object or something that can be converted into a result object.
        """
        ...

    def on_error(
        self, query: Query, error: Exception
    ) -> SearchHandlerCallbackReturns:
        r"""|coro|

        Override this function to add an error response behavior to this handler's callback.

        If the error was handled:
            You can return/yield almost anything, and flogin will convert it into a list of :class:`~flogin.jsonrpc.results.Result` objects before sending it to flow.

        If the error was not handled:
            Return a :class:`~flogin.jsonrpc.responses.ErrorResponse` object

        Parameters
        ----------
        query: :class:`~flogin.query.Query`
            The query that was being handled when the error occured.
        error: :class:`Exception`
            The error that occured

        Returns
        -------
        :class:`~flogin.jsonrpc.responses.ErrorResponse` | list[:class:`~flogin.jsonrpc.results.Result`] | :class:`~flogin.jsonrpc.results.Result` | str | Any
            A list of results, an results, or something that can be converted into a list of results.

        Yields
        ------
        :class:`~flogin.jsonrpc.results.Result` | str | Any
            A result object or something that can be converted into a result object.
        """
        ...

    if not TYPE_CHECKING:
        @copy_doc(callback)
        async def callback(self, query: Query):
            raise RuntimeError("Callback was not overriden")

        @copy_doc(on_error)
        async def on_error(self, query: Query, error: Exception):
            LOG.exception(
                f"Ignoring exception in reuslt callback ({self!r})", exc_info=error
            )
            return ErrorResponse.internal_error(error)

    @property
    def name(self) -> str:
        """:class:`str`: The name of the search handler's callback"""
        return self.callback.__name__

    def error(self, func: ErrorHandlerT) -> ErrorHandlerT:
        """A decorator that registers a error handler for this search handler.

        For more information see :class:`~flogin.search_handler.SearchHandler.on_error`

        Example
        ---------

        .. code-block:: python3

            @plugin.search()
            async def my_hander(query):
                ..

            @my_handler.error
            async def my_error_handler(query, error):
                ...

        """

        self.on_error = func  # type: ignore
        return func



from __future__ import annotations

from typing import TYPE_CHECKING, Any, ParamSpec

from .fuzzy_search import FuzzySearchResult
from .plugin_metadata import PluginMetadata

ATS = ParamSpec("ATS")

if TYPE_CHECKING:
    from ..jsonrpc import ExecuteResponse, JsonRPCClient, Result

__all__ = ("FlowLauncherAPI",)


class FlowLauncherAPI:
    r"""This class is a wrapper around Flow's API to make it easy to make requests and receive results.

    .. NOTE::
        Do not initialize this class yourself, instead use :class:`~flogin.plugin.Plugin`'s :attr:`~flogin.plugin.Plugin.api` attribute to get an instance.
    """

    def __init__(self, jsonrpc: JsonRPCClient):
        self.jsonrpc = jsonrpc

    async def __call__(self, method: str, *args: Any, **kwargs: Any) -> ExecuteResponse:
        from ..jsonrpc import ExecuteResponse

        await getattr(self, method)(*args, **kwargs)
        return ExecuteResponse()

    async def fuzzy_search(
        self, text: str, text_to_compare_it_to: str
    ) -> FuzzySearchResult:
        r"""|coro|

        Asks flow how similiar two strings are.

        Parameters
        --------
        text: :class:`str`
            The text
        text_to_compare_it_to: :class:`str`
            The text you want to compare the other text to

        Returns
        --------
        :class:`~flogin.flow_api.fuzzy_search.FuzzySearchResult`
        """

        from ..jsonrpc import ErrorResponse  # circular import

        res = await self.jsonrpc.request("FuzzySearch", [text, text_to_compare_it_to])
        assert not isinstance(res, ErrorResponse)
        return FuzzySearchResult(res.data)

    async def change_query(self, new_query: str, requery: bool = False) -> None:
        r"""|coro|

        Change the query in flow launcher's menu.

        Parameters
        --------
        new_query: :class:`str`
            The new query to change it to
        requery: :class:`bool`
            Whether or not to re-send a query request in the event that the `new_query` is the same as the current query

        Returns
        --------
        None
        """

        from ..jsonrpc import ErrorResponse  # circular import

        res = await self.jsonrpc.request("ChangeQuery", [new_query, requery])
        assert not isinstance(res, ErrorResponse)

    async def show_error_message(self, title: str, text: str) -> None:
        r"""|coro|

        Triggers an error message in the form of a windows notification

        Parameters
        --------
        title: :class:`str`
            The title of the notification
        text: :class:`str`
            The content of the notification

        Returns
        --------
        None
        """

        from ..jsonrpc import ErrorResponse  # circular import

        res = await self.jsonrpc.request("ShowMsgError", [title, text])
        assert not isinstance(res, ErrorResponse)

    async def show_notification(
        self,
        title: str,
        content: str,
        icon: str = "",
        use_main_window_as_owner: bool = True,
    ) -> None:
        r"""|coro|

        Creates a notification window in the bottom right hand of the user's screen

        Parameters
        --------
        title: :class:`str`
            The notification's title
        content: :class:`str`
            The notification's content
        icon: :class:`str`
            The icon to be shown with the notification, defaults to `""`
        use_main_window_as_owner: :class:`bool`
            Whether or not to use the main flow window as the notification's owner. Defaults to `True`

        Returns
        --------
        None
        """

        from ..jsonrpc import ErrorResponse  # circular import

        res = await self.jsonrpc.request(
            "ShowMsg", [title, content, icon, use_main_window_as_owner]
        )
        assert not isinstance(res, ErrorResponse)

    async def open_settings_menu(self) -> None:
        r"""|coro|

        This method tells flow to open up the settings menu.

        Returns
        --------
        None
        """

        from ..jsonrpc import ErrorResponse  # circular import

        res = await self.jsonrpc.request("OpenSettingDialog")
        assert not isinstance(res, ErrorResponse)

    async def open_url(self, url: str, in_private: bool = False) -> None:
        r"""|coro|

        Open up a url in the user's preferred browser, which was set in their Flow Launcher settings.

        Parameters
        --------
        url: :class:`str`
            The url to be opened in the webbrowser
        in_private: :class:`bool`
            Whether or not to open up the url in a private window

        Returns
        --------
        None
        """

        from ..jsonrpc import ErrorResponse  # circular import

        res = await self.jsonrpc.request("OpenUrl", [url, in_private])
        assert not isinstance(res, ErrorResponse)

    async def run_shell_cmd(self, cmd: str, filename: str = "cmd.exe") -> None:
        r"""|coro|

        Tell flow to run a shell command

        Parameters
        --------
        cmd: :class:`str`
            The command to be run
        filename: :class:`str`
            The name of the command prompt instance, defaults to `cmd.exe`

        Returns
        --------
        None
        """

        from ..jsonrpc import ErrorResponse  # circular import

        res = await self.jsonrpc.request("ShellRun", [cmd, filename])
        assert not isinstance(res, ErrorResponse)

    async def restart_flow_launcher(self) -> None:
        r"""|coro|

        This method tells flow launcher to initiate a restart of flow launcher.

        .. WARNING::
            Expect this method to never finish, so clean up and prepare for the plugin to be shut down before calling this.
        """

        from ..jsonrpc import ErrorResponse  # circular import

        res = self.jsonrpc.request("RestartApp")
        assert not isinstance(res, ErrorResponse)

    async def save_all_app_settings(self) -> None:
        r"""|coro|

        This method tells flow to save all app settings.

        Returns
        --------
        None
        """

        from ..jsonrpc import ErrorResponse  # circular import

        res = await self.jsonrpc.request("SaveAppAllSettings")
        assert not isinstance(res, ErrorResponse)

    async def save_plugin_settings(self) -> Any:
        r"""|coro|

        This method tells flow to save plugin settings

        Returns
        --------
        None
        """

        from ..jsonrpc import ErrorResponse  # circular import

        res = await self.jsonrpc.request("SavePluginSettings")
        assert not isinstance(res, ErrorResponse)
        return res.data

    async def reload_all_plugin_data(self) -> None:
        r"""|coro|

        This method tells flow to trigger a reload of all plugins.

        Returns
        --------
        None
        """

        from ..jsonrpc import ErrorResponse  # circular import

        res = await self.jsonrpc.request("ReloadAllPluginDataAsync")
        assert not isinstance(res, ErrorResponse)

    async def show_main_window(self) -> None:
        """|coro|

        This method tells flow to show the main window

        Returns
        --------
        None
        """

        from ..jsonrpc import ErrorResponse  # circular import

        res = await self.jsonrpc.request("ShowMainWindow")
        assert not isinstance(res, ErrorResponse)

    async def hide_main_window(self) -> None:
        r"""|coro|

        This method tells flow to hide the main window

        Returns
        --------
        None
        """

        from ..jsonrpc import ErrorResponse  # circular import

        res = await self.jsonrpc.request("HideMainWindow")
        assert not isinstance(res, ErrorResponse)

    async def is_main_window_visible(self) -> bool:
        r"""|coro|

        This method asks flow if the main window is visible or not

        Returns
        --------
        :class:`bool`
        """

        from ..jsonrpc import ErrorResponse  # circular import

        res = await self.jsonrpc.request("IsMainWindowVisible")
        assert not isinstance(res, ErrorResponse)
        return res.data

    async def check_for_updates(self) -> None:
        r"""|coro|

        This tells flow launcher to check for updates to flow launcher

        .. NOTE::
            This tells flow launcher to check for updates to flow launcher, not your plugin

        Returns
        --------
        None
        """

        from ..jsonrpc import ErrorResponse  # circular import

        res = await self.jsonrpc.request("CheckForNewUpdate")
        assert not isinstance(res, ErrorResponse)

    async def get_all_plugins(self) -> list[PluginMetadata]:
        r"""|coro|

        Get the metadata of all plugins that the user has installed

        Returns
        --------
        list[:class:`~flogin.flow_api.plugin_metadata.PluginMetadata`]
        """

        from ..jsonrpc import ErrorResponse  # circular import

        res = await self.jsonrpc.request("GetAllPlugins")
        assert not isinstance(res, ErrorResponse)
        return [PluginMetadata(plugin["metadata"], self) for plugin in res.data]

    async def add_keyword(self, plugin_id: str, keyword: str) -> None:
        r"""|coro|

        Registers a new keyword for a plugin with flow launcher.

        Parameters
        --------
        plugin_id: :class:`str`
            The id of the plugin that you want the keyword added to
        keyword: :class:`str`
            The keyword to add

        Returns
        --------
        None
        """

        from ..jsonrpc import ErrorResponse  # circular import

        res = await self.jsonrpc.request("AddActionKeyword", [plugin_id, keyword])
        assert not isinstance(res, ErrorResponse)

    async def remove_keyword(self, plugin_id: str, keyword: str) -> None:
        r"""|coro|

        Unregisters a keyword for a plugin with flow launcher.

        Parameters
        --------
        plugin_id: :class:`str`
            The ID of the plugin that you want to remove the keyword from
        keyword: :class:`str`
            The keyword that you want to remove

        Returns
        --------
        None
        """

        from ..jsonrpc import ErrorResponse  # circular import

        res = await self.jsonrpc.request("RemoveActionKeyword", [plugin_id, keyword])
        assert not isinstance(res, ErrorResponse)

    async def open_directory(self, directory: str, file: str | None = None) -> None:
        r"""|coro|

        Opens up a folder in file explorer. If a file is provided, the file will be pre-selected.

        Parameters
        --------
        directory: :class:`str`
            The directory you want to open
        file: Optional[:class:`str`]
            The file in the directory that you want to highlight, defaults to `None`

        Returns
        --------
        None
        """

        from ..jsonrpc import ErrorResponse  # circular import

        res = await self.jsonrpc.request("OpenDirectory", [directory, file])
        assert not isinstance(res, ErrorResponse)

    async def update_results(self, raw_query: str, results: list[Result]) -> None:
        r"""|coro|

        Tells flow to change the results shown to the user

        .. NOTE::
            The ``raw_query`` parameter is required by flow launcher, and must be the same as the current raw query in flow launcher for the results to successfully update.

        Parameters
        ----------
        raw_query: :class:`str`
            Only change the results if the current raw query is the same as this
        results: list[:class:`~flogin.jsonrpc.results.Result`]
            The new results

        Returns
        -------
        None
        """

        from ..jsonrpc import ErrorResponse, QueryResponse  # circular import

        res = await self.jsonrpc.request(
            "UpdateResults", [raw_query, QueryResponse(results).to_dict()]
        )
        assert not isinstance(res, ErrorResponse)



from __future__ import annotations

import asyncio
import json
import logging
import re
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncIterable,
    Awaitable,
    Callable,
    Coroutine,
    Iterable,
    TypeVar,
    TypeVarTuple,
    overload,
)

from .conditions import PlainTextCondition, RegexCondition
from .default_events import get_default_events
from .errors import InvalidContextDataReceived, PluginNotInitialized
from .flow_api.client import FlowLauncherAPI, PluginMetadata
from .jsonrpc import (
    ErrorResponse,
    ExecuteResponse,
    JsonRPCClient,
    QueryResponse,
    Result,
)
from .jsonrpc.responses import BaseResponse
from .query import Query
from .search_handler import SearchHandler
from .settings import Settings
from .utils import MISSING, cached_property, coro_or_gen, setup_logging

if TYPE_CHECKING:
    from ._types import SearchHandlerCallback, SearchHandlerCondition
TS = TypeVarTuple("TS")
EventCallbackT = TypeVar("EventCallbackT", bound=Callable[..., Any])
LOG = logging.getLogger(__name__)

__all__ = ("Plugin",)


class Plugin:
    r"""This class represents your plugin


    Attributes
    --------
    settings: :class:`~flogin.settings.Settings`
        The plugin's settings set by the user
    api: :class:`~flogin.flow_api.client.FlowLauncherAPI`
        An easy way to acess Flow Launcher's API
    """

    def __init__(self) -> None:
        self.jsonrpc: JsonRPCClient = JsonRPCClient(self)
        self.api = FlowLauncherAPI(self.jsonrpc)
        self._metadata: PluginMetadata | None = None
        self._events: dict[str, Callable[..., Awaitable[Any]]] = get_default_events(
            self
        )
        self._search_handlers: list[SearchHandler] = []
        self._results: dict[str, Result] = {}
        self._settings_are_populated: bool = False

    @cached_property
    def settings(self) -> Settings:
        fp = f"../../Settings/Plugins/{self.metadata.name}/Settings.json"
        with open(fp, "r") as f:
            data = json.load(f)
        self._settings_are_populated = True
        LOG.debug(f"Settings filled from file: {data!r}")
        return Settings(data)

    async def _run_event(
        self,
        coro: Callable[..., Awaitable[Any]],
        event_name: str,
        args: Iterable[Any],
        kwargs: dict[str, Any],
        error_handler: Callable[[Exception], Coroutine[Any, Any, Any]] | str = MISSING,
    ) -> Any:
        try:
            return await coro(*args, **kwargs)
        except asyncio.CancelledError:
            pass
        except Exception as e:
            if error_handler is MISSING:
                error_handler = "on_error"
            if isinstance(error_handler, str):
                return await self._events[error_handler](event_name, e, *args, **kwargs)
            else:
                return await error_handler(e)

    def _schedule_event(
        self,
        coro: Callable[..., Awaitable[Any]],
        event_name: str,
        args: Iterable[Any] = MISSING,
        kwargs: dict[str, Any] = MISSING,
        error_handler: Callable[[Exception], Coroutine[Any, Any, Any]] | str = MISSING,
    ) -> asyncio.Task:
        wrapped = self._run_event(
            coro, event_name, args or [], kwargs or {}, error_handler
        )
        return asyncio.create_task(wrapped, name=f"flogin: {event_name}")

    def dispatch(
        self, event: str, *args: Any, **kwargs: Any
    ) -> None | asyncio.Task[None | BaseResponse]:
        method = f"on_{event}"

        # Special Event Cases
        replacements = {
            # "on_context_menu": "_context_menu_wrapper",
            "on_initialize": "_initialize_wrapper",
        }
        method = replacements.get(method, method)

        LOG.debug("Dispatching event %s", method)

        event_callback = self._events.get(method)
        if event_callback:
            return self._schedule_event(event_callback, method, args, kwargs)

    async def _coro_or_gen_to_results(
        self, coro: Awaitable | AsyncIterable
    ) -> list[Result] | ErrorResponse:
        results = []
        raw_results = await coro_or_gen(coro)
        if isinstance(raw_results, ErrorResponse):
            return raw_results
        if isinstance(raw_results, dict):
            res = Result.from_dict(raw_results)
            self._results[res.slug] = res
            results.append(res)
        else:
            if not isinstance(raw_results, list):
                raw_results = [raw_results]
            for raw_res in raw_results:
                res = Result.from_anything(raw_res)
                self._results[res.slug] = res
                results.append(res)
        return results

    async def _initialize_wrapper(self, arg: dict[str, Any]) -> ExecuteResponse:
        LOG.info(f"Initialize: {json.dumps(arg)}")
        self._metadata = PluginMetadata(arg["currentPluginMetadata"], self.api)
        self.dispatch("initialization")
        return ExecuteResponse(hide=False)

    async def process_context_menus(
        self, data: list[Any]
    ) -> QueryResponse | ErrorResponse:
        LOG.debug(f"Context Menu Handler: {data=}")

        if not data:
            raise InvalidContextDataReceived()

        result = self._results.get(data[0])

        if result is not None:
            result.plugin = self
            task = self._schedule_event(
                self._coro_or_gen_to_results,
                event_name=f"ContextMenu-{result.slug}",
                args=[result.context_menu()],
                error_handler=lambda e: self._coro_or_gen_to_results(
                    result.on_context_menu_error(e)
                ),
            )
            results = await task
        else:
            results = []

        if isinstance(results, ErrorResponse):
            return results
        return QueryResponse(results, self.settings._changes)

    async def process_search_handlers(
        self, query: Query
    ) -> QueryResponse | ErrorResponse:
        results = []
        for handler in self._search_handlers:
            if handler.condition(query):
                handler.plugin = self
                task = self._schedule_event(
                    self._coro_or_gen_to_results,
                    event_name=f"SearchHandler-{handler.name}",
                    args=[handler.callback(query)],
                    error_handler=lambda e: self._coro_or_gen_to_results(
                        handler.on_error(query, e)
                    ),
                )
                results = await task
                break

        if isinstance(results, ErrorResponse):
            return results
        return QueryResponse(results, self.settings._changes)

    @property
    def metadata(self) -> PluginMetadata:
        """
        Returns the plugin's metadata.

        Raises
        --------
        :class:`~flogin.errors.PluginNotInitialized`
            This gets raised if the plugin hasn't been initialized yet
        """
        if self._metadata:
            return self._metadata
        raise PluginNotInitialized()

    async def start(self):
        r"""|coro|

        The default startup/setup method. This can be overriden for advanced startup behavior, but make sure to run ``await super().start()`` to actually start your plugin.
        """

        import aioconsole

        reader, writer = await aioconsole.get_standard_streams()
        await self.jsonrpc.start_listening(reader, writer)

    def run(self, *, setup_default_log_handler: bool = True) -> None:
        r"""The default runner. This runs the :func:`~flogin.plugin.Plugin.start` coroutine, and setups up logging.

        Parameters
        --------
        setup_default_log_handler: :class:`bool`
            Whether to setup the default log handler or not, defaults to `True`.
        """

        if setup_default_log_handler:
            setup_logging()

        asyncio.run(self.start())

    def register_search_handler(self, handler: SearchHandler) -> None:
        r"""Register a new search handler

        See the :ref:`search handler section <search_handlers>` for more information about using search handlers.

        Parameters
        -----------
        handler: :class:`~flogin.search_handler.SearchHandler`
            The search handler to be registered
        """

        self._search_handlers.append(handler)
        LOG.info(f"Registered search handler: {handler}")

    def event(self, callback: EventCallbackT) -> EventCallbackT:
        """A decorator that registers an event to listen for.

        All events must be a :ref:`coroutine <coroutine>`.

        .. NOTE::
            See the :ref:`event reference <events>` to see what valid events there are.

        Example
        ---------

        .. code-block:: python3

            @plugin.event
            async def on_initialization():
                print('Ready!')

        """

        name = callback.__name__
        self._events[name] = callback
        return callback

    @overload
    def search(
        self, condition: SearchHandlerCondition
    ) -> Callable[[SearchHandlerCallback], SearchHandler]: ...

    @overload
    def search(
        self, *, text: str
    ) -> Callable[[SearchHandlerCallback], SearchHandler]: ...

    @overload
    def search(
        self, *, pattern: re.Pattern
    ) -> Callable[[SearchHandlerCallback], SearchHandler]: ...

    @overload
    def search(
        self,
    ) -> Callable[[SearchHandlerCallback], SearchHandler]: ...

    def search(
        self,
        condition: SearchHandlerCondition | None = None,
        *,
        text: str = MISSING,
        pattern: re.Pattern = MISSING,
    ) -> Callable[[SearchHandlerCallback], SearchHandler]:
        """A decorator that registers a search handler.

        All search handlers must be a :ref:`coroutine <coroutine>`. See the :ref:`search handler section <search_handlers>` for more information about using search handlers.

        Parameters
        ----------
        condition: Optional[:ref:`condition <condition_example>`]
            The condition to determine which queries this handler should run on. If given, this should be the only argument given.
        text: Optional[:class:`str`]
            A kwarg to quickly add a :class:`~flogin.conditions.PlainTextCondition`. If given, this should be the only argument given.
        pattern: Optional[:class:`re.Pattern`]
            A kwarg to quickly add a :class:`~flogin.conditions.RegexCondition`. If given, this should be the only argument given.

        Example
        ---------

        .. code-block:: python3

            @plugin.on_search()
            async def example_search_handler(data: Query):
                return "This is a result!"

        """

        if condition is None:
            if text is not MISSING:
                condition = PlainTextCondition(text)
            elif pattern is not MISSING:
                condition = RegexCondition(pattern)

        def inner(func: SearchHandlerCallback) -> SearchHandler:
            handler = SearchHandler(condition)
            handler.callback = func  # type: ignore # type is the same
            self.register_search_handler(handler)
            return handler

        return inner



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, Awaitable, Callable

from .jsonrpc import ErrorResponse
from .query import Query
from .settings import Settings

if TYPE_CHECKING:
    from .plugin import Plugin


LOG = logging.getLogger(__name__)

__all__ = ("on_error",)


async def on_error(
    event_method: str, error: Exception, *args: Any, **kwargs: Any
) -> ErrorResponse:
    """gets called when an error occurs in an event"""
    LOG.exception(f"Ignoring exception in event {event_method!r}", exc_info=error)
    return ErrorResponse.internal_error(error)


def get_default_events(plugin: Plugin) -> dict[str, Callable[..., Awaitable[Any]]]:
    def on_query(data: dict[str, Any], raw_settings: dict[str, Any]):
        query = Query(data)
        if plugin._settings_are_populated is False:
            plugin.settings = Settings(raw_settings)
        else:
            plugin.settings._update(raw_settings)
        return plugin.process_search_handlers(query)

    def on_context_menu(data: list[str]):
        return plugin.process_context_menus(data)

    return {
        event.__name__: event
        for event in (
            on_error,
            on_query,
            on_context_menu,
            plugin._initialize_wrapper,
        )
    }
